import json
import pytest
from slubfind.parser import (
    AppDetails, AppSearch, JsonLdResponse, JsonLdDetails, JsonLdSearch
)


def _make(cls, data):
    """Build a parser instance from a Python object (serialised to JSON)."""
    return cls(json.dumps(data))


# ---------------------------------------------------------------------------
# AppDetails
# ---------------------------------------------------------------------------

def test_app_details_ok_and_found_with_valid_id():
    result = _make(AppDetails, {"id": "0-1132486122"})
    assert result.ok is True
    assert result.found is True


def test_app_details_ok_but_not_found_empty_id():
    result = _make(AppDetails, {"id": ""})
    assert result.ok is True
    assert result.found is False


def test_app_details_ok_but_not_found_whitespace_id():
    result = _make(AppDetails, {"id": "   "})
    assert result.ok is True
    assert result.found is False


def test_app_details_not_ok_missing_id():
    result = _make(AppDetails, {"title": "something"})
    assert result.ok is False
    assert result.found is False


def test_app_details_not_ok_non_dict_response():
    result = _make(AppDetails, ["item1", "item2"])
    assert result.ok is False
    assert result.found is False


def test_app_details_not_ok_null_response():
    result = AppDetails("null")
    assert result.ok is False
    assert result.found is False


def test_app_details_ok_but_not_found_numeric_id():
    result = _make(AppDetails, {"id": 42})
    assert result.ok is True
    assert result.found is False


def test_app_details_invalid_json():
    result = AppDetails("not-json{{{")
    assert result.raw is None
    assert result.ok is False
    assert result.found is False


def test_app_details_id_field():
    result = _make(AppDetails, {"id": "0-1132486122"})
    assert result.id == "0-1132486122"


def test_app_details_record_property():
    result = _make(AppDetails, {
        "id": "0-123",
        "record": {"title": "Test Title", "format": "Book"}})
    assert result.record == {"title": "Test Title", "format": "Book"}
    assert result.title == "Test Title"
    assert result.format == "Book"


def test_app_details_record_none_when_missing():
    result = _make(AppDetails, {"id": "0-123"})
    assert result.record is None
    assert result.title is None


def test_app_details_unescape_html_entities():
    result = _make(AppDetails, {
        "id": "0-123",
        "record": {"title": "M&uuml;nchen &amp; Berlin"}})
    assert result.record["title"] == "München & Berlin"
    assert result.title == "München & Berlin"


def test_app_details_convenience_properties():
    result = _make(AppDetails, {
        "id": "0-123",
        "record": {
            "title": "A Title",
            "format": "Book",
            "contributor": "Author Name",
            "identifier": "ISBN-123"}})
    assert result.contributor == "Author Name"
    assert result.identifier == "ISBN-123"


# ---------------------------------------------------------------------------
# AppSearch
# ---------------------------------------------------------------------------

def test_app_search_ok():
    result = _make(AppSearch, {
        "numFound": 5, "start": 0, "docs": [{"id": "1"}], "facets": {}})
    assert result.ok is True
    assert result.num_found == 5
    assert result.start == 0
    assert result.docs == [{"id": "1"}]


def test_app_search_ok_minimal():
    result = _make(AppSearch, {"docs": []})
    assert result.ok is True
    assert result.num_found == 0
    assert result.start == 0
    assert result.docs == []


def test_app_search_not_ok_missing_docs():
    result = _make(AppSearch, {"numFound": 0})
    assert result.ok is False
    assert result.num_found == 0
    assert result.docs == []


def test_app_search_not_ok_non_dict():
    result = _make(AppSearch, ["item"])
    assert result.ok is False


def test_app_search_invalid_json():
    result = AppSearch("not-json{{{")
    assert result.raw is None
    assert result.ok is False
    assert result.num_found == 0
    assert result.docs == []


# ---------------------------------------------------------------------------
# JsonLdResponse
# ---------------------------------------------------------------------------

def test_jsonld_ok_with_graph():
    result = _make(JsonLdResponse, {
        "@context": {"so": "http://schema.org/"},
        "@graph": [
            {"@id": "https://example.com/id/0-123",
             "so:name": "Test", "@type": "so:Book"}
        ]})
    assert result.ok is True
    assert len(result.graph) == 1
    assert result.graph[0]["@type"] == "so:Book"


def test_jsonld_ok_empty_graph():
    result = _make(JsonLdResponse, {
        "@context": {"so": "http://schema.org/"}, "@graph": []})
    assert result.ok is True
    assert result.graph == []


def test_jsonld_ok_no_graph_key():
    result = _make(JsonLdResponse, {
        "@context": {"so": "http://schema.org/"}})
    assert result.ok is True
    assert result.graph == []


def test_jsonld_not_ok_missing_context():
    result = _make(JsonLdResponse, {"@graph": []})
    assert result.ok is False
    assert result.graph == []


def test_jsonld_not_ok_non_dict():
    result = _make(JsonLdResponse, ["item"])
    assert result.ok is False


def test_jsonld_invalid_json():
    result = JsonLdResponse("not-json{{{")
    assert result.raw is None
    assert result.ok is False
    assert result.graph == []


# ---------------------------------------------------------------------------
# JsonLdDetails
# ---------------------------------------------------------------------------

def test_jsonld_details_found_with_real_id():
    result = _make(JsonLdDetails, {
        "@context": {"so": "http://schema.org/"},
        "@graph": [
            {"@id": "https://katalog.slub-dresden.de/id/0-123",
             "so:name": "Test Book", "so:author": "Author",
             "so:url": "https://example.com", "@type": "so:Book"}
        ]})
    assert result.ok is True
    assert result.found is True
    assert result.id == "https://katalog.slub-dresden.de/id/0-123"
    assert result.name == "Test Book"
    assert result.author == "Author"
    assert result.url == "https://example.com"
    assert result.type == "so:Book"


def test_jsonld_details_not_found_bare_id():
    result = _make(JsonLdDetails, {
        "@context": {"so": "http://schema.org/"},
        "@graph": [
            {"@id": "https://katalog.slub-dresden.de/id/",
             "so:name": "Not Found"}
        ]})
    assert result.ok is True
    assert result.found is False


def test_jsonld_details_not_found_bare_id_no_trailing_slash():
    result = _make(JsonLdDetails, {
        "@context": {"so": "http://schema.org/"},
        "@graph": [
            {"@id": "https://katalog.slub-dresden.de/id",
             "so:name": "Not Found"}
        ]})
    assert result.ok is True
    assert result.found is False


def test_jsonld_details_not_found_empty_graph():
    result = _make(JsonLdDetails, {
        "@context": {"so": "http://schema.org/"},
        "@graph": []})
    assert result.ok is True
    assert result.found is False


def test_jsonld_details_properties_empty_graph():
    result = _make(JsonLdDetails, {
        "@context": {"so": "http://schema.org/"},
        "@graph": []})
    assert result.id is None
    assert result.name is None
    assert result.author is None
    assert result.url is None
    assert result.type is None


# ---------------------------------------------------------------------------
# JsonLdSearch
# ---------------------------------------------------------------------------

def test_jsonld_search_ok():
    result = _make(JsonLdSearch, {
        "@context": {"so": "http://schema.org/"},
        "@graph": [
            {"@id": "https://example.com/1", "so:name": "Result 1"},
            {"@id": "https://example.com/2", "so:name": "Result 2"}
        ]})
    assert result.ok is True
    assert len(result.graph) == 2


def test_jsonld_search_not_ok():
    result = _make(JsonLdSearch, {"data": "invalid"})
    assert result.ok is False
    assert result.graph == []



"""
parser module of ``slubfind``
"""
from txpyfind.parser import JSONResponse


class HoldingStatus(JSONResponse):  # pylint: disable=R0903
    """
    Parser for holding status responses.

    Expected structure::

        {
            "status": ...,
            "location": "...",
            "links": [...]
        }
    """

    def __init__(self, plain):
        super().__init__(plain)
        self.ok = isinstance(self.raw, dict)

    @property
    def status(self):
        """Return the status field."""
        return self.raw.get("status") if self.ok else None

    @property
    def location(self):
        """Return the location field."""
        return self.raw.get("location") if self.ok else None

    @property
    def links(self):
        """Return the links field."""
        return self.raw.get("links") if self.ok else None


class AppDetails(JSONResponse):  # pylint: disable=R0903
    """
    Parser for app-format detail view responses.

    Expected structure::

        {
            "record": {"format": ..., "title": ..., ...},
            "id": "0-1132486122",
            "oa": 0,
            "thumbnail": "...",
            "copies": [...],
            "parts": {...}
        }
    """

    def __init__(self, plain):
        super().__init__(plain)
        self.ok = isinstance(self.raw, dict) and "id" in self.raw
        self.found = self.ok and isinstance(
            self.raw["id"], str) and len(self.raw["id"].strip()) > 0
        self.id = self._field("id")

    def _unescape_dict(self, d):
        """Apply inherited _unescape() to each value of a dict."""
        if not isinstance(d, dict):
            return d
        return {k: self._unescape(v) for k, v in d.items()}

    @property
    def record(self):
        """Return the unescaped record dict."""
        raw_record = self.raw.get("record") if self.ok else None
        if isinstance(raw_record, dict):
            return self._unescape_dict(raw_record)
        return None

    @property
    def title(self):
        """Return the title from the record dict."""
        rec = self.record
        return rec.get("title") if rec else None

    @property
    def format(self):
        """Return the format from the record dict."""
        rec = self.record
        return rec.get("format") if rec else None

    @property
    def contributor(self):
        """Return the contributor from the record dict."""
        rec = self.record
        return rec.get("contributor") if rec else None

    @property
    def identifier(self):
        """Return the identifier from the record dict."""
        rec = self.record
        return rec.get("identifier") if rec else None


class AppSearch(JSONResponse):  # pylint: disable=R0903
    """
    Parser for app-format search view responses.

    Expected structure::

        {
            "numFound": 42,
            "start": 0,
            "docs": [...],
            "facets": {...}
        }
    """

    def __init__(self, plain):
        super().__init__(plain)
        self.ok = isinstance(self.raw, dict) and "docs" in self.raw
        self.num_found = self.raw.get("numFound", 0) if self.ok else 0
        self.start = self.raw.get("start", 0) if self.ok else 0
        self.docs = self.raw.get("docs", []) if self.ok else []


class JsonLdResponse(JSONResponse):  # pylint: disable=R0903
    """
    Parser for JSON-LD responses (detail and search views).

    Expected structure::

        {
            "@context": {"so": "http://schema.org/"},
            "@graph": [
                {
                    "@id": "https://katalog.slub-dresden.de/id/0-...",
                    "so:name": "...",
                    "so:author": "...",
                    "so:url": "...",
                    "@type": "so:Book"
                }
            ]
        }
    """

    def __init__(self, plain):
        super().__init__(plain)
        self.ok = isinstance(self.raw, dict) and "@context" in self.raw
        self.graph = self.raw.get("@graph", []) if self.ok else []


class JsonLdDetails(JsonLdResponse):  # pylint: disable=R0903
    """Parser for JSON-LD detail view responses."""

    def __init__(self, plain):
        super().__init__(plain)
        self.found = self.ok and len(self.graph) > 0 and not (
            self.graph[0].get("@id", "").rstrip("/").endswith("/id"))

    @property
    def id(self):
        """Return the @id of the first graph entry."""
        return self.graph[0].get("@id") if self.graph else None

    @property
    def name(self):
        """Return the so:name of the first graph entry."""
        return self.graph[0].get("so:name") if self.graph else None

    @property
    def author(self):
        """Return the so:author of the first graph entry."""
        return self.graph[0].get("so:author") if self.graph else None

    @property
    def url(self):
        """Return the so:url of the first graph entry."""
        return self.graph[0].get("so:url") if self.graph else None

    @property
    def type(self):
        """Return the @type of the first graph entry."""
        return self.graph[0].get("@type") if self.graph else None


class JsonLdSearch(JsonLdResponse):  # pylint: disable=R0903
    """Parser for JSON-LD search view responses."""
    pass



import argparse
import io
import pytest
from unittest.mock import MagicMock, patch

from slubfind.cli import (
    build_parser,
    json_dumps,
    main,
    merge_facets,
    parse_facet,
)


# ---------------------------------------------------------------------------
# json_dumps
# ---------------------------------------------------------------------------

def test_json_dumps_default_compact():
    result = json_dumps({"a": 1})
    assert result == '{"a":1}'


def test_json_dumps_pretty():
    result = json_dumps({"a": 1}, pretty=True)
    assert '"a": 1' in result
    assert "\n" in result


def test_json_dumps_non_ascii():
    result = json_dumps({"k": "ä"})
    assert "ä" in result


# ---------------------------------------------------------------------------
# parse_facet
# ---------------------------------------------------------------------------

def test_parse_facet_valid():
    key, val = parse_facet("language=ger")
    assert key == "language"
    assert val == "ger"


def test_parse_facet_no_equals():
    with pytest.raises(argparse.ArgumentTypeError, match="KEY=VALUE"):
        parse_facet("language")


def test_parse_facet_unknown_key():
    with pytest.raises(argparse.ArgumentTypeError, match="unknown facet key"):
        parse_facet("unknown_key=value")


def test_parse_facet_valid_facet_avail():
    key, val = parse_facet("facet_avail=Online")
    assert key == "facet_avail"
    assert val == "Online"


def test_parse_facet_invalid_facet_avail():
    with pytest.raises(argparse.ArgumentTypeError, match="unknown value"):
        parse_facet("facet_avail=Invalid")


def test_parse_facet_unconstrained_accepts_any():
    key, val = parse_facet("language=ger")
    assert key == "language"
    assert val == "ger"


def test_parse_facet_all_format_de14_values():
    from slubfind.client import SlubFind
    for val in SlubFind.FACET_VALUES["format_de14"]:
        key, v = parse_facet(f"format_de14={val}")
        assert key == "format_de14"
        assert v == val


def test_parse_facet_invalid_format_de14():
    with pytest.raises(argparse.ArgumentTypeError, match="unknown value"):
        parse_facet("format_de14=Nonexistent")


def test_parse_facet_valid_access_state():
    key, val = parse_facet("access_state=open")
    assert key == "access_state"
    assert val == "open"


def test_parse_facet_invalid_access_state():
    with pytest.raises(argparse.ArgumentTypeError, match="unknown value"):
        parse_facet("access_state=invalid")


def test_parse_facet_value_with_equals():
    # value portion may itself contain '='
    key, val = parse_facet("language=a=b")
    assert key == "language"
    assert val == "a=b"


# ---------------------------------------------------------------------------
# merge_facets
# ---------------------------------------------------------------------------

def test_merge_facets_empty():
    assert merge_facets([]) is None
    assert merge_facets(None) is None


def test_merge_facets_single():
    assert merge_facets([("language", "ger")]) == [{"language": "ger"}]


def test_merge_facets_multiple():
    result = merge_facets([("language", "ger"), ("format_de14", "Buch")])
    assert result == [{"language": "ger"}, {"format_de14": "Buch"}]


def test_merge_facets_duplicate_key():
    result = merge_facets([("language", "ger"), ("language", "eng")])
    assert result == [{"language": "ger"}, {"language": "eng"}]


# ---------------------------------------------------------------------------
# helpers
# ---------------------------------------------------------------------------

EXAMPLE_URL = ("https://katalog.slub-dresden.de/"
               "?tx_find_find%5Bq%5D%5Bdefault%5D=manfred+bonitz")


def _make_result(raw):
    """Build a mock result object with a .raw attribute."""
    m = MagicMock()
    m.raw = raw
    return m


def _mock_url_parser(is_ok=True):
    """Build a mock URLParser result."""
    p = MagicMock()
    p.is_ok = is_ok
    p.query = "manfred bonitz"
    p.qtype = "default"
    p.facets = None
    p.page = 0
    p.count = 0
    p.sort = ""
    return p


def _run(argv, find_mock, capsys):
    """Run main(), injecting find_mock via make_find patch. Returns exit code."""
    with patch("slubfind.cli.make_find", return_value=find_mock):
        with pytest.raises(SystemExit) as exc_info:
            main(argv)
    return exc_info.value.code


# ---------------------------------------------------------------------------
# cmd_query
# ---------------------------------------------------------------------------

def test_cmd_query_success(capsys):
    find = MagicMock()
    find.get_query.return_value = _make_result({"docs": []})
    code = _run(["query", "python"], find, capsys)
    assert code == 0
    assert '{"docs"' in capsys.readouterr().out


def test_cmd_query_pretty(capsys):
    find = MagicMock()
    find.get_query.return_value = _make_result({"docs": []})
    code = _run(["--pretty", "query", "python"], find, capsys)
    assert code == 0
    assert "\n" in capsys.readouterr().out


def test_cmd_query_show_url(capsys):
    find = MagicMock()
    find.url_query.return_value = "https://example.com/q"
    code = _run(["--show-url", "query", "python"], find, capsys)
    assert code == 0
    assert "https://example.com/q" in capsys.readouterr().out


def test_cmd_query_error(capsys):
    find = MagicMock()
    find.get_query.return_value = None
    code = _run(["query", "python"], find, capsys)
    assert code == 1


def test_cmd_query_with_facet(capsys):
    find = MagicMock()
    find.get_query.return_value = _make_result({"docs": []})
    code = _run(["query", "--facet", "language=ger", "python"], find, capsys)
    assert code == 0
    _, kwargs = find.get_query.call_args
    assert kwargs["facet"] == [{"language": "ger"}]


def test_cmd_query_raw_solr_response(capsys):
    find = MagicMock()
    find.get_query.return_value = _make_result({"response": {"docs": []}})
    code = _run(
        ["--export-format", "raw-solr-response", "query", "python"],
        find, capsys)
    assert code == 0


# ---------------------------------------------------------------------------
# cmd_document
# ---------------------------------------------------------------------------

def test_cmd_document_success(capsys):
    find = MagicMock()
    find.get_document.return_value = _make_result({"id": "0-123"})
    code = _run(["document", "0-123"], find, capsys)
    assert code == 0
    assert '"id"' in capsys.readouterr().out


def test_cmd_document_show_url(capsys):
    find = MagicMock()
    find.url_document.return_value = "https://example.com/doc/0-123"
    code = _run(["--show-url", "document", "0-123"], find, capsys)
    assert code == 0
    assert "https://example.com/doc/0-123" in capsys.readouterr().out


def test_cmd_document_not_found(capsys):
    find = MagicMock()
    find.get_document.return_value = None
    code = _run(["document", "0-123"], find, capsys)
    assert code == 1


def test_cmd_document_show_url_none(capsys):
    find = MagicMock()
    find.url_document.return_value = None
    code = _run(["--show-url", "document", "0-123"], find, capsys)
    assert code == 1


# ---------------------------------------------------------------------------
# cmd_scroll
# ---------------------------------------------------------------------------

def test_cmd_scroll_success(capsys):
    find = MagicMock()
    find.scroll_get_query.return_value = [{"id": "1"}, {"id": "2"}]
    code = _run(["scroll", "python"], find, capsys)
    assert code == 0
    assert '"id"' in capsys.readouterr().out


def test_cmd_scroll_stream(capsys):
    find = MagicMock()
    find.stream_get_query.return_value = iter([{"id": "1"}, {"id": "2"}])
    code = _run(["scroll", "--stream", "python"], find, capsys)
    assert code == 0
    lines = [l for l in capsys.readouterr().out.splitlines() if l.strip()]
    assert len(lines) == 2


def test_cmd_scroll_show_url(capsys):
    find = MagicMock()
    find.url_query.return_value = "https://example.com/scroll"
    code = _run(["--show-url", "scroll", "python"], find, capsys)
    assert code == 0
    assert "https://example.com/scroll" in capsys.readouterr().out


def test_cmd_scroll_error(capsys):
    find = MagicMock()
    find.scroll_get_query.return_value = None
    code = _run(["scroll", "python"], find, capsys)
    assert code == 1


# ---------------------------------------------------------------------------
# cmd_settings
# ---------------------------------------------------------------------------

def test_cmd_settings_success(capsys):
    find = MagicMock()
    find.settings.return_value = {"key": "value"}
    code = _run(["settings"], find, capsys)
    assert code == 0
    assert '"key"' in capsys.readouterr().out


def test_cmd_settings_error(capsys):
    find = MagicMock()
    find.settings.return_value = None
    code = _run(["settings"], find, capsys)
    assert code == 1


# ---------------------------------------------------------------------------
# cmd_solr_params
# ---------------------------------------------------------------------------

def test_cmd_solr_params_success(capsys):
    find = MagicMock()
    find.solr_params.return_value = {"params": {"q": "python"}}
    code = _run(["solr-params", "python"], find, capsys)
    assert code == 0
    assert '"params"' in capsys.readouterr().out


def test_cmd_solr_params_error(capsys):
    find = MagicMock()
    find.solr_params.return_value = None
    code = _run(["solr-params", "python"], find, capsys)
    assert code == 1


# ---------------------------------------------------------------------------
# cmd_solr_request
# ---------------------------------------------------------------------------

def test_cmd_solr_request_success(capsys):
    find = MagicMock()
    find.solr_request.return_value = "https://solr.example.com/select?q=python"
    code = _run(["solr-request", "python"], find, capsys)
    assert code == 0
    assert "https://solr.example.com" in capsys.readouterr().out


def test_cmd_solr_request_error(capsys):
    find = MagicMock()
    find.solr_request.return_value = None
    code = _run(["solr-request", "python"], find, capsys)
    assert code == 1


# ---------------------------------------------------------------------------
# --from-url
# ---------------------------------------------------------------------------

def test_query_from_url(capsys):
    find = MagicMock()
    find.url_parser.return_value = _mock_url_parser()
    find.get_query.return_value = _make_result({"docs": []})
    code = _run(["query", "--from-url", EXAMPLE_URL], find, capsys)
    assert code == 0
    assert '"docs"' in capsys.readouterr().out


def test_query_from_url_invalid(capsys):
    find = MagicMock()
    find.url_parser.return_value = _mock_url_parser(is_ok=False)
    code = _run(["query", "--from-url", EXAMPLE_URL], find, capsys)
    assert code == 1


def test_query_from_url_show_url(capsys):
    find = MagicMock()
    find.url_parser.return_value = _mock_url_parser()
    find.url_query.return_value = "https://example.com/q"
    code = _run(
        ["--show-url", "query", "--from-url", EXAMPLE_URL], find, capsys)
    assert code == 0
    assert "https://example.com/q" in capsys.readouterr().out


def test_query_no_query_no_from_url(capsys):
    find = MagicMock()
    code = _run(["query"], find, capsys)
    assert code == 1


def test_scroll_from_url(capsys):
    find = MagicMock()
    find.url_parser.return_value = _mock_url_parser()
    find.scroll_get_query.return_value = [{"id": "1"}]
    code = _run(["scroll", "--from-url", EXAMPLE_URL], find, capsys)
    assert code == 0
    assert '"id"' in capsys.readouterr().out


def test_solr_params_from_url(capsys):
    find = MagicMock()
    find.url_parser.return_value = _mock_url_parser()
    find.solr_params.return_value = {"params": {"q": "manfred bonitz"}}
    code = _run(["solr-params", "--from-url", EXAMPLE_URL], find, capsys)
    assert code == 0
    assert '"params"' in capsys.readouterr().out


def test_solr_request_from_url(capsys):
    find = MagicMock()
    find.url_parser.return_value = _mock_url_parser()
    find.solr_request.return_value = "https://solr.example.com/select?q=test"
    code = _run(["solr-request", "--from-url", EXAMPLE_URL], find, capsys)
    assert code == 0
    assert "https://solr.example.com" in capsys.readouterr().out


# ---------------------------------------------------------------------------
# main — no command
# ---------------------------------------------------------------------------

def test_main_no_command():
    with pytest.raises(SystemExit) as exc_info:
        main([])
    assert exc_info.value.code == 1



from slubfind.client import SlubFind


def test_get_document():
    slub_find = SlubFind()
    result = slub_find.app_document("0-1132486122")
    assert result is not None
    assert isinstance(result.raw, dict)
    assert "id" in result.raw


def test_get_holding_status_document():
    slub_find = SlubFind()
    result = slub_find.holding_status_document("0-1132486122")
    assert result is not None
    assert isinstance(result.raw, dict)


def test_get_query():
    slub_find = SlubFind()
    result = slub_find.app_search("manfred bonitz")
    assert result is not None
    assert isinstance(result.raw, dict)


def test_settings():
    slub_find = SlubFind()
    result = slub_find.settings()
    assert result is not None
    assert isinstance(result, dict)


def test_solr_params():
    slub_find = SlubFind()
    result = slub_find.solr_params("manfred bonitz")
    assert result is not None
    assert isinstance(result, dict)


def test_solr_params_via_url():
    slub_find = SlubFind()
    result = slub_find.solr_params_via_url(
        "https://katalog.slub-dresden.de/"
        "?tx_find_find%5Bq%5D%5Bdefault%5D=manfred+bonitz")
    assert result is not None
    assert isinstance(result, dict)


def test_get_query_raw_solr_response():
    slub_find = SlubFind(export_format="raw-solr-response")
    result = slub_find.get_query("manfred bonitz")
    assert result is not None
    assert isinstance(result.raw, dict)
    assert "response" in result.raw


def test_solr_request():
    slub_find = SlubFind()
    result = slub_find.solr_request("manfred bonitz")
    assert result is not None
    assert isinstance(result, str)
    assert result.startswith("http")



"""
client module of ``slubfind`` package
"""
import re
from txpyfind import utils
from txpyfind.client import Find
from txpyfind.parser import JSONResponse

from .parser import (
    AppDetails, AppSearch, HoldingStatus, JsonLdDetails, JsonLdSearch
)


class SlubFind(Find):
    """
    ``SlubFind`` class from ``slubfind.client`` module
    """

    EXPORT_FORMATS = [
        "app",
        "json-ld",
        "json-holding-status",
        "json-all",
        "json-solr-params",
        "json-solr-request",
        "raw-solr-response"
    ]

    QUERY_TYPES = [
        "default",
        "author",
        "title",
        "topic",
        "barcode",
        "ident",
        "rvk_facet",
        "signatur",
        "imprint",
        "series2",
        "provenance"
    ]

    FACET_VALUES = {
        "facet_avail": [
            "Online", "Local", "Free"
        ],
        "format_de14": [
            "Article, E-Article", "Audio", "Book, E-Book", "Braille",
            "Collection", "Electronic Resource",
            "Electronic Resource (Data Carrier)",
            "Electronic Resource (Remote Access)",
            "Journal, E-Journal", "Kit", "Manuscript", "Map",
            "Microform", "Musical Score", "Norm", "Standard",
            "Physical Object", "Proceeding", "Software", "Thesis",
            "Unknown Format", "Video", "Visual Media", "Website"
        ],
        "access_state": [
            "open", "embargoed", "restricted", "metadata",
            "closed", "unknown"
        ],
    }

    FACETS = [
        "facet_avail",
        "format_de14",
        "publishDateSort",
        "branch_collcode",
        "branch",
        "license",
        "access_state",
        "language",
        "thema",
        "author",
        "facet_music_notation_de14",
        "music_heading_browse",
        "provenance",
        "mega_collection"
    ]

    def __init__(
            self,
            base_url="https://katalog.slub-dresden.de",
            export_format="app",
            export_page=1369315142,
            parser_class=JSONResponse):
        super().__init__(
            base_url,
            document_path="id",
            query_types=self.QUERY_TYPES,
            facets=self.FACETS,
            count_limit=1000,
            sort_pattern=re.compile(
                r"^(score|publishDateSort|title_sort|id)[+ ](asc|desc)$"),
            export_format=export_format,
            export_page=export_page,
            parser_class=parser_class)

    def app_document(
            self,
            document_id,
            type_num=None):
        """
        fetch detail view in app format
        """
        return self.get_document(
            document_id,
            data_format="app",
            type_num=type_num,
            parser_class=AppDetails)

    def holding_status_document(
            self,
            document_id,
            type_num=None):
        """
        fetch holding status for a document
        """
        return self.get_document(
            document_id,
            data_format="json-holding-status",
            type_num=type_num,
            parser_class=HoldingStatus)

    def app_search(  # pylint: disable=R0913,R0917
            self,
            query,
            qtype="default",
            facet=None,
            page=0,
            count=0,
            sort="",
            type_num=None,
            parser_class=None):
        """
        fetch query view in app format
        """
        return self.get_query(
            query,
            qtype=qtype,
            facet=facet,
            page=page,
            count=count,
            sort=sort,
            data_format="app",
            type_num=type_num,
            parser_class=parser_class or AppSearch)

    def jsonld_document(
            self,
            document_id,
            type_num=None):
        """
        fetch detail view in JSON-LD format
        """
        return self.get_document(
            document_id,
            data_format="json-ld",
            type_num=type_num,
            parser_class=JsonLdDetails)

    def jsonld_search(  # pylint: disable=R0913,R0917
            self,
            query,
            qtype="default",
            facet=None,
            page=0,
            count=0,
            sort="",
            type_num=None,
            parser_class=None):
        """
        fetch query view in JSON-LD format
        """
        return self.get_query(
            query,
            qtype=qtype,
            facet=facet,
            page=page,
            count=count,
            sort=sort,
            data_format="json-ld",
            type_num=type_num,
            parser_class=parser_class or JsonLdSearch)

    def settings(self):
        """
        get the settings used for TYPO3-find
        """
        url = self.url_query(
            "qqqqq",
            qtype="default",
            facet=None,
            page=0,
            count=0,
            sort="",
            data_format="json-all",
            type_num=None)
        response = utils.json_request(url)
        if isinstance(response, dict) and "settings" in response:
            return response["settings"]
        return None

    def solr_params(  # pylint: disable=R0913,R0917
            self,
            query,
            qtype="default",
            facet=None,
            page=0,
            count=0,
            sort="",
            type_num=None):
        """
        get the parameters used for the Solr request
        """
        url = self.url_query(
            query=query,
            qtype=qtype,
            facet=facet,
            page=page,
            count=count,
            sort=sort,
            data_format="json-solr-params",
            type_num=type_num)
        url = utils.add_tx_param(url, "omitHeader", "false")
        response = utils.json_request(url)
        if isinstance(response, dict):
            return response
        return None

    def solr_params_via_url(
            self,
            url,
            type_num=None):
        """
        get the parameters used for the Solr request via given TYPO3-find URL
        """
        url = self.url_parser(url)
        if url.is_ok:
            return self.solr_params(
                url.query,
                qtype=url.qtype,
                facet=url.facets,
                page=url.page,
                count=url.count,
                sort=url.sort,
                type_num=type_num)
        return None

    def solr_request(  # pylint: disable=R0913,R0917
            self,
            query,
            qtype="default",
            facet=None,
            page=0,
            count=0,
            sort="",
            type_num=None):
        """
        get the URL of the Solr request
        """
        url = self.url_query(
            query=query,
            qtype=qtype,
            facet=facet,
            page=page,
            count=count,
            sort=sort,
            data_format="json-solr-request",
            type_num=type_num)
        url = utils.add_tx_param(url, "debug", "1")
        response = utils.json_request(url)
        if isinstance(response, dict) and "url" in response:
            return response["url"]
        return None
